/*
This file is part of libmsppath
Copyright © 2006  Mikko Rasa, Mikkosoft Productions
Distributed under the LGPL
*/
#include <msp/strings/utils.h>
#include "path.h"
#include "utils.h"

using namespace std;

namespace Msp {
namespace Path {

/**
Returns the number of components in the path.
*/
unsigned Path::size() const
{
	if(!path.size()) return 0;
	if(path.size()==1 && path[0]==DIRCHAR) return 1;
	unsigned count=1;
	for(string::const_iterator i=path.begin(); i!=path.end(); ++i)
		if(*i==DIRCHAR) ++count;
	return count;
}

bool Path::is_absolute() const
{
#ifdef WIN32
	if(is_windows_drive((*this)[0])) return true;
#endif
	if(path[0]==DIRCHAR) return true;
	return false;
}

/**
Extracts the given component range from the path.
*/
Path Path::subpath(unsigned start, unsigned count) const
{
	Path result;
	iterator i=begin();
	for(unsigned j=0; (j<start && i!=end()); ++j)
		++i;
	for(unsigned j=0; (j<count && i!=end()); ++j)
	{
		result/=*i;
		++i;
	}
	return result;
}

/**
Attaches another path to the end of this one.  An absolute path replaces the
existing data.  ".." elements annihilate the last component and "." elements
are ignored.
*/
Path &Path::operator/=(const Path &p)
{
	if(p.is_absolute())
		path=p.path;
	else
	{
		for(iterator i=p.begin(); i!=p.end(); ++i)
			add_component(*i);
	}
	return *this;
}

/**
Returns the path component at the specified index.  Negative indices count from
the end of the path.
*/
string Path::operator[](int n) const
{
	if(n>=0)
	{
		for(iterator i=begin(); i!=end(); ++i,--n)
			if(!n) return *i;
	}
	else
	{
		for(iterator i=--end();; --i)
		{
			if(!++n) return *i;
			if(i==begin()) break;
		}
	}

	return "";
}

bool Path::operator==(const Path &p) const
{
#ifdef WIN32
	return !strcasecmp(path, p.path);
#else
	return path==p.path;
#endif
}

void Path::init(const string &p)
{
	unsigned start=0;
	if(p[0]=='/' || p[0]=='\\')
		add_component(string(1, DIRCHAR));
	while(1)
	{
		unsigned slash=p.find_first_of("/\\",start);
		if(slash>start)
		{
			add_component(p.substr(start, slash-start));
		}
		if(slash==string::npos)
			break;
		start=slash+1;
	}
}

/**
Adds a single component to the path, emulating the cd command.
*/
void Path::add_component(const string &comp)
{
	if(comp.empty())
		;
	else if(comp.size()==1 && comp[0]==DIRCHAR)
	{
		// Replace the path with the root directory
#ifdef WIN32
		unsigned slash=path.find(DIRCHAR);
		if(is_windows_drive(path.substr(0, slash)))
			path=path.substr(0,2);
		else
#endif
		path=comp;
	}
#ifdef WIN32
	else if(is_windows_drive(comp))
		path=comp;
#endif
	else if(comp=="..")
	{
		if(path.empty())
			path=comp;
		// .. in root directory is a no-op
		else if(path.size()==1 && path[0]==DIRCHAR)
			;
#ifdef WIN32
		else if(is_windows_drive(path))
			;
#endif
		else
		{
			unsigned slash=path.rfind(DIRCHAR);
			unsigned start=(slash==string::npos)?0:slash+1;
			if(!path.compare(start, string::npos, ".."))
			{
				// If the last component already is a .., add another
				path+=DIRCHAR;
				path+=comp;
			}
			else
				// Otherwise, erase the last component
				path.erase(slash, string::npos);
		}
	}
	else if(comp!="." || path.empty())
	{
		if(path==".")
			path="";
		if(path.size()>1 || (path.size()==1 && path[0]!=DIRCHAR))
			path+=DIRCHAR;
		path+=comp;
	}
}

Path::iterator::iterator(const Path &p):
	path(p),
	start(0)
{
	if(path.path.empty())
		start=end=string::npos;
	else if(path.path[0]==DIRCHAR)
		end=1;
#ifdef WIN32
	else if(path.path.size()>2 && path.path[2]==DIRCHAR && is_windows_drive(path.path.substr(0,2)))
		end=2;
#endif
	else
		end=path.path.find(DIRCHAR);
}

Path::iterator &Path::iterator::operator++()
{
	start=end;
	if(start>=path.path.size()) return *this;
	if(path.path[start]==DIRCHAR) ++start;
	end=path.path.find(DIRCHAR,start);
	return *this;
}

Path::iterator &Path::iterator::operator--()
{
	end=start;
	if(end==0) return *this;
	if(end>1 && end<path.path.size() && path.path[end]!=DIRCHAR) --end;
	start=path.path.rfind(DIRCHAR,end-1);
	if(start==string::npos)
		start=0;
	else if(start<end-1)
		++start;
	return *this;
}

string Path::iterator::operator*() const
{
	if(start>=path.path.size()) return "";
	if(start==end) return "";
	return path.path.substr(start,end-start);
}

} // namespace Path
} // namespace Msp
